#ifndef __cplusplus
#error Must use C++ for the type tapfloat.
#endif


#if !defined(__APFLOAT_H)
#include "apfloat.h"
#endif

#if !defined(__TAPFLOAT_H)
#define __TAPFLOAT_H


// Low-level transformed apfloat functions

apstruct *transform (apstruct *s, int m, size_t ssize, size_t n);
apstruct *tapmul (apstruct *a, apstruct *b1, apstruct *b2, apstruct *b3);
apstruct *tapmul (apstruct *a1, apstruct *a2, apstruct *a3, apstruct *b1, apstruct *b2, apstruct *b3);
apstruct *tapadd (apstruct *a, apstruct *b, int m);


// The tapfloat class
// Includes three pointers to apstructs
class tapfloat
{
public:
    // Constructors
    tapfloat ();
    tapfloat (apstruct *d1, apstruct *d2, apstruct *d3);
    tapfloat (tapfloat &d);                     // NOT const &d

    // Destructor
    ~tapfloat ();

    // Operator functions
    friend tapfloat operator+ (tapfloat &, tapfloat &);
    friend apfloat operator* (apfloat &, tapfloat &);
    friend apfloat operator* (tapfloat &, apfloat &);
    friend apfloat operator* (tapfloat &, tapfloat &);
    tapfloat &operator+= (tapfloat &);
    tapfloat &operator= (tapfloat &);

    // Other functions
    void unique (void);

    // Implementation
    apstruct *ap1, *ap2, *ap3;
};


inline tapfloat::tapfloat ()
{
    ap1 = 0;
    ap2 = 0;
    ap3 = 0;
}


inline tapfloat::tapfloat (apstruct *d1, apstruct *d2, apstruct *d3)
{
    ap1 = d1;
    ap2 = d2;
    ap3 = d3;
}


// Transformed apfloat functions

tapfloat transform (apfloat x, size_t ssize = (size_t) DEFAULT, size_t padsize = (size_t) DEFAULT);
apfloat pow (tapfloat base, unsigned long exp);
apfloat pow (tapfloat base, unsigned exp);
apfloat pow (tapfloat base, long exp);
apfloat pow (tapfloat base, int exp);


#endif  // __TAPFLOAT_H
